import { Component, ElementRef, EventEmitter, HostListener, Input, OnChanges, OnDestroy, OnInit, Output, ViewChild, ChangeDetectorRef } from '@angular/core';
import { Subject } from 'rxjs';
import { debounceTime } from 'rxjs/operators';
import { EditorItem, HotZoneImageItem, Label, MySprite } from './Unit';
import TWEEN from '@tweenjs/tween.js';
import * as uuid from 'uuid';

@Component({
  selector: 'app-custom-hot-zone',
  templateUrl: './custom-hot-zone.component.html',
  styleUrls: ['./custom-hot-zone.component.scss']
})
export class CustomHotZoneComponent implements OnInit, OnDestroy {

  _bgItem = null;

  @Input()
  set bgItem(v) {
    this._bgItem = v;
    this.init();
  }
  get bgItem() {
    return this._bgItem;
  }

  @Input() hotZoneItemArr = null;
  @Input() zoneWidth: number = 960;
  @Input() zoneHeight: number = 155;
  @Input() title: string = "选区";
  @Input() assetsRoot: string = "";
  @Input() defaultBGImage: string = "";
  @Input() defaultItemBG: string = "";
  @Input() itemWidth: number = 50;
  @Input() itemHeight: number = 50;
  @Input() fontSize: string = "48";
  @Input() fontName: string = "FuturaBT-Medium";
  @Input() fontColor: string = "#000";
  @Input() customBackground: boolean = true;
  @Input() enableResize: boolean = true;
  @Output() save = new EventEmitter();


  @ViewChild('canvas', {static: true }) canvas: ElementRef;
  @ViewChild('wrap', {static: true }) wrap: ElementRef;
  @HostListener('window:resize', ['$event'])
  onResize(event) {
    this.g_winResizeEventStream.next();
  }

  g_winResizeEventStream = new Subject();
  firstTouch = true;
  canvasWidth: number;
  canvasHeight: number;

  canvasBaseW: number;
  canvasBaseH: number;
  mapScale = 1;
  hotZoneArr = null;
  ctx;
  mx;
  my; // 点击坐标

  // 声音
  bgAudio = new Audio();
  images = new Map();
  animationId: any;
  // winResizeEventStream = new Subject();
  canvasLeft;
  canvasTop;
  renderArr;
  imgArr = [];
  oldPos;
  curItem;
  bg: MySprite;
  changeSizeFlag = false;
  changeTopSizeFlag = false;
  changeRightSizeFlag = false;
  hotZoneChanged = false;
  scale = 1;

  constructor(private el: ElementRef, private changeDetectorRef: ChangeDetectorRef) {

  }

  ngOnInit() {
    this.canvasWidth = Number(this.zoneWidth);
    this.canvasHeight = Number(this.zoneHeight);
    this.canvasBaseW = Number(this.zoneWidth);
    this.canvasBaseH = Number(this.zoneHeight);

    if(!this.bgItem.url){
      this.bgItem.url = this.getURL(this.defaultBGImage)
    } else {
      this.bgItem.url = this.getURL(this.bgItem.url)
    }
    
    this.init();
    this.initListener();
    this.update();
  }

  ngOnDestroy() {
    window.cancelAnimationFrame(this.animationId);
  }

  onBackgroundUploadSuccess(e) {
    this.bgItem.url = e.url;
    this.refreshBackground(() => {
      this.autoSave()
    });
  }

  refreshBackground(callBack = null) {
    if (!this.bg) {
      this.bg = new MySprite(this.ctx);
      this.renderArr.push(this.bg);
    }
    const bg = this.bg;
    
    if (this.getURL(this.bgItem.url)) {
      bg.load(this.getURL(this.bgItem.url)).then(() => {
        const rate1 = this.canvasWidth / bg.width;
        const rate2 = this.canvasHeight / bg.height;
        const rate = Math.min(rate1, rate2);
        bg.setScaleXY(rate);
        bg.x = this.canvasWidth / 2;
        bg.y = this.canvasHeight / 2;
        if (callBack) {
          callBack();
        }
      }).catch((error) => {
        console.log(error)
      })
    }
  }

  addHotZoo() {
    const item = this.getHotZoneItem();
    this.hotZoneArr.push(item);
    this.refreshHotZoneId();
    this.autoSave()
  }

  deleteHotZoo(i) {
    this.hotZoneArr.splice(i, 1);
    this.refreshHotZoneId();
    this.autoSave()
  }

  setTextByUID(text, uid) {
    let mached = false;
    this.hotZoneArr.forEach(element => {
      if(element.uid == uid) {
        element.text = text
        element.label.text = text
        mached = true;
      }
    });
    if(mached) {
      this.autoSave()
    }
  }

  setAudioByUID(audio_url, uid) {
    let mached = false;
    this.hotZoneArr.forEach(element => {
      if(element.uid == uid) {
        element.audio_url = audio_url;
        mached = true;
      }
    });
    if(mached) {
      this.autoSave()
    }
  }

  refreshImage(img) {
    this.hideAllLineDash();
    img.picItem = this.getPicItem(img);
    this.refreshImageId();
  }

  refreshHotZoneId() {
    for (let i = 0; i < this.hotZoneArr.length; i++) {
      this.hotZoneArr[i].index = i;
    }
  }

  refreshHotZoneText(item) {
    item.initNew()
  }

  refreshImageId() {
    for (let i = 0; i < this.imgArr.length; i++) {
      this.imgArr[i].id = i;
      if (this.imgArr[i].picItem) {
        this.imgArr[i].picItem.text = 'Image-' + (i + 1);
      }
    }
  }

  getHotZoneItem(saveData = null, newRect?) {
    const itemW = this.itemWidth;
    const itemH = this.itemHeight;
    const item = new HotZoneImageItem(this.ctx);
    let imageUrl = ""
    if (saveData) {
      item.initFont(this.fontSize, this.fontName, this.fontColor)
      imageUrl = this.getURL(saveData.media.image_url);
      item.init(imageUrl, saveData.media.text, saveData.media.type, (w, h) => {
        const saveRect = saveData.rect;
        item.scaleX = (saveData.rect.width / saveData.scale) / w;
        item.scaleY = (saveData.rect.height / saveData.scale) / h;
        item.x = saveRect.x / saveData.scale + newRect.x
        item.y = saveRect.y / saveData.scale + newRect.y
      });
      item.uid = saveData.uid
      item.image_url = imageUrl
      item.text = saveData.media.text?saveData.media.text:""
      item.type = saveData.media.type
      item.audio_url = saveData.media.audio_url?saveData.media.audio_url:""
    } else {
      item.uid = uuid.v4()
      imageUrl = this.getURL(this.defaultItemBG);
      item.initFont(this.fontSize, this.fontName, this.fontColor)
      item.init(imageUrl, "", "Text", () => {
        item.x = this.canvasWidth / 2 - this.itemWidth /  2;
        item.y = this.canvasHeight / 2 - 20;
      }, () => this.autoSave());
    }
    item.anchorX = 0.5;
    item.anchorY = 0.5;
    return item;
  }

  getPicItem(img, saveData = null) {
    const item = new EditorItem(this.ctx);
    item.load(img.pic_url).then(img => {
      let maxW, maxH;
      if (this.bg) {
        maxW = this.bg.width * this.bg.scaleX;
        maxH = this.bg.height * this.bg.scaleY;
      } else {
        maxW = this.canvasWidth;
        maxH = this.canvasHeight;
      }
      let scaleX = maxW / 3 / item.width;
      let scaleY = maxH / 3 / item.height;
      if (item.height * scaleX < this.canvasHeight) {
        item.setScaleXY(scaleX);
      } else {
        item.setScaleXY(scaleY);
      }
      item.x = this.canvasWidth / 2;
      item.y = this.canvasHeight / 2;
      if (saveData) {
        const saveRect = saveData.rect;
        item.setScaleXY(saveRect.width / item.width);
        item.x = saveRect.x + saveRect.width / 2;
        item.y = saveRect.y + saveRect.height / 2;
      } else {
        item.showLineDash();
      }
      this.autoSave()
    })
    return item;
  }

  init() {
    this.initData();
    this.initCtx();
    this.initItem();
  }

  initItem() {
    if (!this.bgItem) {
      this.bgItem = { };
    } else {
      if(!this.bgItem.questions) {
        this.bgItem.questions = []
      }
      this.refreshBackground(() => {
        if (!this.hotZoneItemArr) {
          this.hotZoneItemArr = [];
        } else {
          this.initHotZoneArr();
        }
      });
    }
  }

  initHotZoneArr() {
    let curBgRect;
    if (this.bg) {
      curBgRect = this.bg.getBoundingBox();
    } else {
      curBgRect = { x: 0, y: 0, width: this.canvasWidth, height: this.canvasHeight };
    }
    let oldBgRect = this.bgItem.rect;
    if (!oldBgRect) {
      oldBgRect = curBgRect;
    }
    const rate = curBgRect.width / oldBgRect.width;
    this.hotZoneArr = [];
    const arr = this.hotZoneItemArr.concat();
    for (let i = 0; i < arr.length; i++) {
      const data = JSON.parse(JSON.stringify(arr[i]));
      data.rect.x *= rate;
      data.rect.y *= rate;
      data.rect.width *= rate;
      data.rect.height *= rate;
      // data.rect.x += curBgRect.x;
      // data.rect.y += curBgRect.y;
      const item = this.getHotZoneItem(data, { x: curBgRect.x, y: curBgRect.y });
      this.hotZoneArr.push(item);
    }
    this.refreshHotZoneId();
  }

  initData() {
    this.canvasWidth = this.wrap.nativeElement.clientWidth;
    this.canvasHeight = this.wrap.nativeElement.clientWidth * (this.zoneHeight / this.zoneWidth);
    this.scale = this.zoneWidth * 0.95 / this.wrap.nativeElement.clientWidth
    this.mapScale = this.canvasWidth / this.canvasBaseW;
    this.renderArr = [];
    this.bg = null;
    this.imgArr = [];
    this.hotZoneArr = [];
  }

  initCtx() {
    this.ctx = this.canvas.nativeElement.getContext('2d');
    this.canvas.nativeElement.width = this.canvasWidth;
    this.canvas.nativeElement.height = this.canvasHeight;
  }
  
  mapDown(event) {
    this.oldPos = { x: this.mx, y: this.my };
    const arr = this.hotZoneArr;
    for (let i = arr.length - 1; i >= 0; i--) {
      const item = arr[i];
      if (item) {
        if (this.checkClickTarget(item)) {
          if (item.lineDashFlag && this.checkClickTarget(item.arrow)) {
            if(!this.enableResize){
              return
            }
            this.changeItemSize(item);
          } else if (item.lineDashFlag && this.checkClickTarget(item.arrowTop)) {
            if(!this.enableResize){
              return
            }
            this.changeItemTopSize(item);
          } else if (item.lineDashFlag && this.checkClickTarget(item.arrowRight)) {
            if(!this.enableResize){
              return
            }
            this.changeItemRightSize(item);
          } else {
            this.changeCurItem(item);
          }
          this.hotZoneChanged = true;
          this.changeDetectorRef.markForCheck();
          this.changeDetectorRef.detectChanges();
          return;
        }

      }
    }
  }

  mapMove(event) {
    if (!this.curItem) { return; }

    if (this.changeSizeFlag) {
      this.changeSize();

    } else if (this.changeTopSizeFlag) {
      this.changeTopSize();

    } else if (this.changeRightSizeFlag) {
      this.changeRightSize();
    } else {
      let addX = this.mx - this.oldPos.x;
      let addY = this.my - this.oldPos.y;
      this.curItem.x += addX;
      this.curItem.y += addY;
    }

    this.oldPos = { x: this.mx, y: this.my };
  }

  mapUp(event) {
    this.curItem = null;
    this.changeSizeFlag = false;
    this.changeTopSizeFlag = false;
    this.changeRightSizeFlag = false;
  }

  changeSize() {
    const rect = this.curItem.getBoundingBox();
    let distance = this.my - rect.y;
    let lenW = this.mx - rect.x;
    let lenH = rect.height - distance
    let minLen = 20;
    let sx, sy;
    if (lenW < minLen) {
      lenW = minLen;
    }
    sx = lenW / this.curItem.width;
    if (lenH < minLen) {
      lenH = minLen;
    }
    sy = lenH / this.curItem.height;
    this.curItem.y = this.curItem.y + distance
    this.curItem.scaleX = sx;
    this.curItem.scaleY = sy;
  }

  changeTopSize() {
    const rect = this.curItem.getBoundingBox();
    let lenH = rect.y - this.my + rect.height;
    this.curItem.y = this.my
    let minLen = 20;
    let s;
    if (lenH < minLen) {
      lenH = minLen;
    }
    s = lenH / this.curItem.height;
    this.curItem.setScaleXY(s);
  }

  changeRightSize() {
    const rect = this.curItem.getBoundingBox();
    let lenW = this.mx - rect.x;
    let minLen = 20;
    let s;
    if (lenW < minLen) {
      lenW = minLen;
    }
    s = lenW / this.curItem.width;
    this.curItem.setScaleXY(s);
  }

  changeItemSize(item) {
    this.curItem = item;
    this.changeSizeFlag = true;
  }

  changeItemTopSize(item) {
    this.curItem = item;
    this.changeTopSizeFlag = true;
  }

  changeItemRightSize(item) {
    this.curItem = item;
    this.changeRightSizeFlag = true;
  }

  changeCurItem(item) {
    this.hideAllLineDash();
    this.curItem = item;
    this.curItem.showLineDash();
  }

  handlePartChange(e, QI, index) {
    let a = JSON.parse(JSON.stringify(this.bgItem.questions[QI].combin))
    a[index] = Number(e)
    this.bgItem.questions[QI].combin = a
    this.autoSave()
  }

  hideAllLineDash() {
    for (let i = 0; i < this.imgArr.length; i++) {
      if (this.imgArr[i].picItem) {
        this.imgArr[i].picItem.hideLineDash();
      }
    }
  }

  update() {
    this.animationId = window.requestAnimationFrame(this.update.bind(this));
    // 清除画布内容
    this.ctx.clearRect(0, 0, this.canvasWidth, this.canvasHeight);
    for (let i = 0; i < this.renderArr.length; i++) {
      this.renderArr[i].update(this);
    }
    this.updateArr(this.hotZoneArr);
    TWEEN.update();
  }

  updateArr(arr) {
    if (arr) {
      for (let i = 0; i < arr.length; i++) {
        arr[i].update();
      }
    }
  }

  renderAfterResize() {
    this.canvasWidth = this.wrap.nativeElement.clientWidth;
    this.canvasHeight = this.wrap.nativeElement.clientHeight;
    this.init();
  }

  initListener() {
    const element = this.canvas.nativeElement;
    this.g_winResizeEventStream.pipe(debounceTime(500)).subscribe(data => {
      this.renderAfterResize();
    });
    
    const addTouchListener = () => {
      element.addEventListener('touchstart', touchDownFunc);
      element.addEventListener('touchmove', touchMoveFunc);
      element.addEventListener('touchend', touchUpFunc);
      element.addEventListener('touchcancel', touchUpFunc);
    };
    const removeTouchListener = () => {
      element.removeEventListener('touchstart', touchDownFunc);
      element.removeEventListener('touchmove', touchMoveFunc);
      element.removeEventListener('touchend', touchUpFunc);
      element.removeEventListener('touchcancel', touchUpFunc);
    };

    const addMouseListener = () => {
      element.addEventListener('mousedown', mouseDownFunc);
      element.addEventListener('mousemove', mouseMoveFunc);
      element.addEventListener('mouseup', mouseUpFunc);
    };
    const removeMouseListener = () => {
      element.removeEventListener('mousedown', mouseDownFunc);
      element.removeEventListener('mousemove', mouseMoveFunc);
      element.removeEventListener('mouseup', mouseUpFunc);
    };

    const touchDownFunc = (e) => {
      if (this.firstTouch) {
        this.firstTouch = false;
        removeMouseListener();
      }
      setMxMyByTouch(e);
      this.mapDown(e);
    };
    const touchMoveFunc = (e) => {
      setMxMyByTouch(e);
      this.mapMove(e);
    };
    const touchUpFunc = (e) => {
      setMxMyByTouch(e);
      this.mapUp(e);
    };

    const mouseDownFunc = (e) => {
      if (this.firstTouch) {
        this.firstTouch = false;
        removeTouchListener();
      }
      setMxMyByMouse(e);
      this.mapDown(e);
    };
    const mouseMoveFunc = (e) => {
      setMxMyByMouse(e);
      this.mapMove(e);
    };
    const mouseUpFunc = (e) => {
      setMxMyByMouse(e);
      this.mapUp(e);
    };

    const setMxMyByTouch = event => {
      if (event.touches.length <= 0) {
        return;
      }
      if (this.canvasLeft == null) {
        setParentOffset();
      }
      this.mx = event.touches[0].pageX - this.canvasLeft;
      this.my = event.touches[0].pageY - this.canvasTop;
    };

    const setParentOffset = () => {
      const rect = this.canvas.nativeElement.getBoundingClientRect();
      this.canvasLeft = rect.left;
      this.canvasTop = rect.top;
    };

    const setMxMyByMouse = (event) => {
      this.mx = event.offsetX;
      this.my = event.offsetY;
    };

    addMouseListener();
    addTouchListener();

  }

  IsPC() {
    if (window['ELECTRON']) {
      return false; // 封装客户端标记
    }
    if (document.body.ontouchstart !== undefined) {
      return false;
    } else {
      return true;
    }
  }


  checkClickTarget(target) {
    const rect = target.getBoundingBox();
    if (this.checkPointInRect(this.mx, this.my, rect)) {
      return true;
    }
    return false;
  }

  checkPointInRect(x, y, rect) {
    if (x >= rect.x && x <= rect.x + rect.width) {
      if (y >= rect.y && y <= rect.y + rect.height) {
        return true;
      }
    }
    return false;
  }

  autoSave() {
    // console.log("Auto save")
    // const bgItem = JSON.parse(JSON.stringify(this.bgItem));
    const bgItem = this.bgItem
    if (this.bg) {
      bgItem['rect'] = this.bg.getBoundingBox();
    } else {
      bgItem['rect'] = { x: 0, y: 0, width: Math.round(this.canvasWidth * 100) / 100, height: Math.round(this.canvasHeight * 100) / 100 };
    }
    if(this.bgItem.url.indexOf("http") == -1) {
      bgItem.url = this.defaultBGImage
    }

    const hotZoneItemArr = [];
    const hotZoneArr = this.hotZoneArr;
    for (let i = 0; i < hotZoneArr.length; i++) {
      const hotZoneItem = {
        index: hotZoneArr[i].index,
      };
      hotZoneItem['rect'] = hotZoneArr[i].getBoundingBox();
      const currentX = hotZoneItem['rect'].x
      const currentY = hotZoneItem['rect'].y
      hotZoneItem['uid'] = hotZoneArr[i].uid
      hotZoneItem['media'] = {}
      hotZoneItem['scale'] = this.scale
      hotZoneItem['rect'].x = (Math.round((currentX - bgItem['rect'].x) * 1000) / 1000) * this.scale;
      hotZoneItem['rect'].y = (Math.round((currentY - bgItem['rect'].y) * 1000) / 1000) * this.scale;
      hotZoneItem['rect'].width = (Math.round((hotZoneItem['rect'].width) * 1000) / 1000) * this.scale;
      hotZoneItem['rect'].height = (Math.round((hotZoneItem['rect'].height) * 1000) / 1000) * this.scale;
      if(hotZoneArr[i].image_url.indexOf("http") == -1) {
        hotZoneItem['media'].image_url = this.defaultItemBG
      } else {
        hotZoneItem['media'].image_url = hotZoneArr[i].image_url
      }
      hotZoneItem['media'].text = hotZoneArr[i].text
      hotZoneItem['media'].type = hotZoneArr[i].type
      hotZoneItem['media'].audio_url = hotZoneArr[i].audio_url ? hotZoneArr[i].audio_url : ""
      hotZoneItemArr.push(hotZoneItem);
    }
    this.save.emit({ bgItem, hotZoneItemArr });
    this.hotZoneChanged = false;
    this.changeDetectorRef.markForCheck();
    this.changeDetectorRef.detectChanges();
  }

  getURL(url: string) {
    if(url && url.indexOf("http")==-1 && url.indexOf("/")==-1){
      return this.assetsRoot.replace(/_PH_IMAGE_NAME_/g, url)
    } else {
      return url
    }
  }

  saveClick() {
    this.autoSave()
  }
}
